"""Models used for different aspects of a Colmena application"""

from colmena.models.results import Result, ResourceRequirements, FailureInformation, SerializationMethod

__all__ = ['Result', 'ResourceRequirements', 'FailureInformation', 'ExecutableTask', 'SerializationMethod']
